import logging
from distutils.util import strtobool
from typing import Optional, Tuple

from ppa import ui

logger = logging.getLogger(f"ppa-tools.{__name__}")


class MissingApprovalParameter(Exception):
    pass


class InvalidResponder(Exception):
    pass


def process_approval_response(response: dict, responder_usernames: Optional[list] = None) -> Tuple[bool, str]:
    """
    Process a the response to an approval email
    generated by mail.create_approval.

    Args:
        response: Dict. Web hook response data.

        responder_usernames: List. Optional. Usernames permitted
            to respond to the request.

    Returns:
        2-tuple (approved (bool), responder (str))

    Raises:
        InvalidResponder
        MissingApprovalParameter
    """
    logger.info(f"response: {response}")
    logger.info(f"responder usernames: {responder_usernames}")
    if responder_usernames and not isinstance(responder_usernames, list):
        responder_usernames = [responder_usernames]

    try:
        approved = strtobool(response["values"]["approve"][0])
    except KeyError:
        message = 'Could not find URL parameter "approve" in approval response.'
        ui.output_error(message)
        raise MissingApprovalParameter(message)

    try:
        responder = response.get("username").split("\\")[1]
    except IndexError:
        responder = response.get("username")

    if responder_usernames and responder not in responder_usernames:
        valid_responders: str = ", ".join(responder_usernames)
        message = (
            f"Request was responded to by {responder}, but "
            "that username is not a valid responder. Valid "
            f"responders to this request are {valid_responders}."
        )
        ui.output_error(message)
        raise InvalidResponder(message)

    return (approved, responder)
