"""
Generate

This module contains actions used to generate table data & passwords.
"""

import uuid

from typing import List, Dict, Any, Union, Optional

from ppa import ui
from ppa_tools import generate


def table_from_dictionaries(
    header: List[str], keys: List[str], dictionaries: List[Dict[str, Any]]
) -> Dict[str, Union[List[str], List[List[str]]]]:
    r"""Generate table data from a list of dictionaries.

!!! question "Minimum Plugin Version: 1.3.0"

??? info "Input"

    - __header:__ a list of table column headers

    - __keys:__ a list of corresponding dictionary keys, one for each `header`

    - __dictionaries:__ a list of dictionaries to get the values of each `key` from

??? success "Output"

    A dictionary containing the keys `header` & `rows`.

??? example

    - Finding failed SSH logons using [extract_values](../ppa_tools.text/#extract_values) & saving them as `failed_logons`

    - Using this action to turn `failed_logons` into table data, & saving it as `failed_logons_table`

    - Displaying a table using the data generated by this action

    ```YAML linenums="1" hl_lines="10 20 21 26 27"
    - ppa_tools.text.extract_values:
        extract:
          - name: username
            expression: for (.*) from
          - name: ip_address
            expression: from (\d+\.\d+\.\d+\.\d+) port
        text: >
          2020-08-11T08:19:40.463248+00:00 machine-name sshd[10349]: Failed password for user_1 from 1.2.3.4 port 58041 ssh2
          2020-08-12T13:24:07.463248+00:00 machine-name sshd[10349]: Failed password for user_2 from 2.3.4.5 port 58041 ssh2
      save: failed_logons

    - ppa_tools.generate.table_from_dictionaries:
        header:
          - Username
          - IP Address
        keys:
          - username
          - ip_address
      load:
        dictionaries: failed_logons
      save: failed_logons_table

    - ppa.ui.output_table:
        text: Failed SSH Logons
      load:
        header: failed_logons_table.header
        rows: failed_logons_table.rows
    ```

!!! tip "Headers & Keys"

    The `header` & `keys` lists must be the same length, & the attributes must be in the same order.

    If a supplied key is missing from a dictionary, its value will be blank.
    """

    def _get_value(d, k):
        return d.get(k) if d.get(k) is not None else ""

    if len(header) != len(keys):
        message = "Cannot generate table data as the supplied 'header' & 'keys' lists are not the same length."
        ui.output_error(message)
        raise ValueError(message)
    return {"header": header, "rows": [[_get_value(d, key) for key in keys] for d in dictionaries]}


def html_table(header: List[str], rows: List[List[str]]) -> str:
    r"""Generate an HTML table to use in a markdown document or email.

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __header:__ a list of table column headers

    - __rows:__ the table contents as a list of lists, where each inner list is a row of values

??? success "Output"

    The generated HTML table as a string.

??? example

    ```YAML linenums="1"
    - ppa_tools.html.generate_table:
        header:
          - City
          - Country
        rows:
          - - Paris
            - France
          - - London
            - England
      save: html_table
    ```
    """
    header_styles = "; ".join(
        ["padding: 5px", "font-size: 13px", "font-weight: heavy", "text-align: center", "border: 1px solid grey"]
    )
    cell_styles = "; ".join(["padding: 5px", "font-size: 12px", "border: 1px solid gray", "text-align: center"])

    row_template = "<tr>\n{row}\n</tr>"
    table_header = row_template.format(row="\n".join([f'<th style="{header_styles}">{head}</th>' for head in header]))
    table_rows = "\n".join(
        [row_template.format(row="\n".join([f'<td style="{cell_styles}">{cell}</td>' for cell in row])) for row in rows]
    )

    return f"""<table style="margin: auto;">
    {table_header}
    {table_rows}
</table>"""


def regex_password(regex: str = r"[A-Za-z0-9\-£\:\$\#\%\&\^]", length: int = 24) -> str:
    r"""Generate a password based on a regular expression.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __regex:__ a regular expression containing a set of characters (defaults to _[A-Za-z0-9\-£\:\$\#\%\&\^]_)

        - __length:__ the length of the generated password (defaults to _15_)

    ??? success "Output"
        The generated password as a string.

    ??? example

        Generating a 20 character password using the default character set.

        ```YAML linenums="1"
        - ppa_tools.generate.regex_password:
            length: 20
          save: generated_password
        ```
    """
    return generate.password_from_regex(regex=regex, length=length)


def dictionary_password(
    *, length: int = 15, word_count: int = 3, valid_word_lengths: Optional[List[int]] = None
) -> Dict[str, str]:
    """Generate a password containing 1 or more dictionary words.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __length:__ the length of the generated password (defaults to _15_)

        - __word_count:__ the number of words the password should consist of (defaults to _3_)

        - __valid_word_lengths:__ a list of valid word lengths (defaults to _3_-_8_)

    ??? success "Output"

        A dictionary containing the generated password in the following cases:

        __lower__: lower case
        __upper__: upper case
        __title__: title case

    ??? example

        Generating a 20 character password consisting of 4 words.

        ```YAML linenums="1"
        - ppa_tools.generate.dictionary_password:
            length: 20
            word_count: 4
          save: generated_password
        ```
    """
    valid_word_lengths = valid_word_lengths or list(range(3, 9))
    generated_password = generate.password_from_words(length=length, word_count=word_count, valid_word_lengths=valid_word_lengths)
    return {
        "lower": generated_password.password_lower,
        "upper": generated_password.password_upper,
        "title": generated_password.password_title,
    }


def dictionary_password_interactive(
    *,
    length: int = 15,
    word_count: int = 3,
    valid_word_lengths: Optional[List[int]] = None,
    case: Optional[str] = None,
    title: Optional[str] = None,
) -> str:
    """Interactively generate a password containing 1 or more dictionary words.

    Each time a password is generated, the Task Operator will be given the choice to:

    - _Use Password_
    - _Generate Another_

    The process can be repeated as many times as needed.

    !!! question "Minimum Plugin Version: 2.0.0"

    !!! danger "Remember"

        Each generated password will be displayed to the Task Operator in a markdown document.

    ??? info "Input"

        - __length:__ the length of the generated password (defaults to `15`)

        - __word_count:__ the number of words the password should consist of (defaults to `3`)

        - __valid_word_lengths:__ a list of valid word lengths (defaults to `3`-`8`)

        - __case:__ the case (`lower`, `upper`, `title`) to output the password in (defaults to `lower`)

        - __title:__ a title displayed to the Task Operator (defaults to `Generated Password`)

    ??? success "Output"

        The generated password in the case chosen by the `case` input.

    ??? example

        Generating an upper case password using the default `word_count` & `length`:

        ```YAML linenums="1"
        - ppa_tools.generate.dictionary_password_interactive:
            case: upper
          save: generated_password
        ```
    """
    valid_word_lengths = valid_word_lengths or list(range(3, 9))
    title = title or "Generated Password"
    case = (case or "lower").lower()

    if case not in {"lower", "upper", "title"}:
        ui.output_error(f"Supplied case '{case}' is not recognised.")
        case = ui.input_choice("Please select a password case", options=["Lower", "Upper", "Title"]).lower()

    markdown_id = str(uuid.uuid4())
    confirm_id = str(uuid.uuid4())
    while True:
        generated_password = generate.password_from_words(
            length=length, word_count=word_count, valid_word_lengths=valid_word_lengths
        )
        password = getattr(generated_password, f"password_{case}")
        ui.output_markdown(
            f"""
### {title}


<html>
    <body>
        <div style="font-weight: bold; float: left;">Password: </div>
        <div style="padding-left: 5px; font-style: bold; float: left;">{password}</div>
        <div style="float: left; padding-left: 10px;">(</div>
            <div style="color: orange; font-style: italic; float: left;">{generated_password.display_title}</div>
        <div style="float: left;">)</div>
    </body>
</html>
        """,
            element_id=markdown_id,
        )
        if ui.input_confirm("Please choose an option", confirm="Use Password", cancel="Generate Another", element_id=confirm_id):
            return password
